<?php

namespace Api\PhpUtils\Http;
/*
 * Copyright 2011-2012 Hipu Holding Inc, Beijing. All Rights Reserved.
 * DO NOT ALERT OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * This code is distributed as free software within the company.
 * You can redistribute it and/or modify it without any restriction.
 * 
 * If you need additional information or have any questions,
 * please contact: xukf@hipu.com.
 */

/**
 * Description of HttpUtil
 *
 * @author mengweifu
 * @date 2021-7-12 20:00:00
 */


use Api\PhpUtils\Log\FileLog;
use Api\PhpUtils\Mon\MonUtil;

class HttpUtil
{
    private const CODE_SUCCESS = 0;
    private const CODE_FAIL = 999999;

    static function HEAD($url, $timeout = 10000, $retries = 1, $headers = false, $proxy = false)
    {
        return self::CURL($url, $timeout, $retries, $headers, false, $proxy, 'HEAD');
    }
    static function GET($url, $params,  $timeout = 10000, $retries = 1, $headers = false, $proxy = false, $curl_opts = array())
    {
        $url = $url ."?". http_build_query($params);
        return self::CURL($url, $timeout, $retries, $headers, [], $proxy, 'GET', $curl_opts);
    }
    
    static function PUT($url, $post, $timeout = 10000, $retries = 1, $headers = false, $proxy = false)
    {
        return self::CURL($url, $timeout, $retries, $headers, $post, $proxy, 'PUT');
    }

    static function POST($url, $post, $timeout = 10000, $retries = 1, $headers = false, $proxy = false, $curl_opts = array())
    {
        return self::CURL($url, $timeout, $retries, $headers, $post, $proxy, 'POST', $curl_opts);
    }

    static function DELETE($url, $timeout = 10000, $retries = 1, $headers = false, $proxy = false)
    {
        return self::CURL($url, $timeout, $retries, $headers, false, $proxy, 'DELETE');
    }
    static function CURL($url, $timeout, $retries, $headers, $post = false, $proxy = false,  $method = 'GET', $curl_opts = array())
    {

        $ch = self::CH($url, $timeout, $headers, $post, $proxy, $method, $curl_opts);

        $response = self::defaultResponse();

        if (is_resource($ch) === true)
        {
            //返回结果
            $result = false;

            while (($result === false) && ($retries-- > 0))
            {
                $result = curl_exec($ch);
            }

            // { todo
            if (curl_errno($ch) !== 0)
            {
                $GLOBALS['ERROR_MESSAGE'] = curl_error($ch);
            }
            $GLOBALS['HTTP_CODE'] = curl_getinfo($ch, CURLINFO_HTTP_CODE);
            if( isset($GLOBALS['ENABLE_HEADER_REQUEST_TIME']) )
            {
                $GLOBALS['BELENDER_CURL_INFO'] = curl_getinfo($ch);
                //在返回header中添加blender响应时间
                $GLOBALS['DEPENDENT-REQUEST-TIME'] = round(curl_getinfo($ch, CURLINFO_TOTAL_TIME),4).'|'.round(curl_getinfo($ch, CURLINFO_NAMELOOKUP_TIME),4).'|'.round(curl_getinfo($ch, CURLINFO_CONNECT_TIME),4);
                $GLOBALS['DEPENDENT-REQUEST'] = array('url'=>$url, 'method'=>$method, 'body'=>$post);
                $GLOBALS['DEPENDENT-URI'] = parse_url($url, PHP_URL_PATH);
                if ($GLOBALS['HTTP_CODE'] === 0)
                {
                    $GLOBALS['DEPENDENT-STATUS'] = 'timeout';
                }
                else if ($GLOBALS['HTTP_CODE'] !== 200)
                {
                    $GLOBALS['DEPENDENT-STATUS'] = '5xx';
                }else
                {
                    $GLOBALS['DEPENDENT-STATUS'] = 'normal';
                }
            }
            unset($GLOBALS['ENABLE_HEADER_REQUEST_TIME']);
            // }
            // 监控
            self::addMon(
                $url,
                $GLOBALS['HTTP_CODE'] ?? -999,
                round(curl_getinfo($ch, CURLINFO_TOTAL_TIME),4) * 1000
            );

            // 日志
            self::addLog($method, $url, $post, $result);
            // 返回结果
            $response = self::createResponse($result, $ch);

            curl_close($ch);
        }

        return $response;
    }

    /**
     *
     * @param $url
     * @param $timeout
     * @param false $headers
     * @param false $post
     * @param false $proxy
     * @param string $method
     * @param array $curl_opts
     * @return false|mixed|resource
     */
    static function CH($url, $timeout, $headers = false, $post = false, $proxy = false,  $method='GET', $curl_opts = array())
    {
        $ch = curl_init($url);

        if (is_resource($ch) === true)
        {
            curl_setopt($ch, CURLOPT_FAILONERROR,           true);
            curl_setopt($ch, CURLOPT_FOLLOWLOCATION,        true);
            curl_setopt($ch, CURLOPT_RETURNTRANSFER,        true);
            curl_setopt($ch, CURLOPT_TCP_NODELAY,           true);
            curl_setopt($ch, CURLOPT_NOSIGNAL,              true);
            curl_setopt($ch, CURLOPT_SSL_VERIFYHOST,        false);
            curl_setopt($ch, CURLOPT_SSL_VERIFYPEER,        false);
            curl_setopt($ch, CURLOPT_BUFFERSIZE,            10240);
            curl_setopt($ch, CURLOPT_TIMEOUT_MS,            $timeout);
            curl_setopt($ch, CURLOPT_CONNECTTIMEOUT_MS,     $timeout);
            curl_setopt($ch, CURLOPT_HTTP_VERSION,          CURL_HTTP_VERSION_1_0);

            if ($headers !== false)
            {
                curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
            }
            if ($proxy !== false)
            {
                curl_setopt($ch, CURLOPT_PROXY, $proxy);
            }
            if (!empty($curl_opts))
            {
                foreach ($curl_opts as $opt => $val)
                {
                    curl_setopt($ch, $opt, $val);
                }
            }
            switch ($method) {
                case 'HEAD':
                    curl_setopt($ch, CURLOPT_CUSTOMREQUEST, 'HEAD');
                    break;
                case 'GET':
                    break;
                case 'POST':
                    curl_setopt($ch, CURLOPT_POST, true);
                    curl_setopt($ch, CURLOPT_POSTFIELDS, $post);
                    break;
                case 'DELETE':
                    curl_setopt($ch, CURLOPT_CUSTOMREQUEST, 'DELETE');
                    break;
                case 'PUT':
                    curl_setopt($ch, CURLOPT_CUSTOMREQUEST, 'PUT');
                    curl_setopt($ch, CURLOPT_POSTFIELDS, $post);
                    break;
                default:
                    die('httpUtil bad method');
                    break;
            }
        }
        return $ch;
    }
    static function MULTI_POST($urls, $posts, $timeouts , $headers, $proxies, $curl_opts)
    {
        $methods  = [];
        foreach ($urls as $k => $url) {
            $methods[$k] = 'POST';
        }
        return self::MULTI_CURL($urls, $methods, $posts, $timeouts , $headers, $proxies, $curl_opts);
    }

    static function MULTI_GET($urls =[],$params = [], $timeouts = [],$headers = [], $proxies = [], $curl_opts = [])
    {
        $methods = $posts = [];
        foreach ($urls as $k => $url) {
            $urls[$k] = $url ."?". http_build_query($params[$k]);
            $methods[$k] = 'GET';
            $posts[$k] = false;
        }

        return self::MULTI_CURL($urls, $methods, $posts, $timeouts , $headers, $proxies, $curl_opts);
    }

    /**
     *
     * @param array $urls 链接
     * @param array $methods 请求方法 (GET POST ...)
     * @param array $posts  POST 内容 参见 https://www.php.net/manual/en/function.curl-setopt.php CURLOPT_POSTFIELDS
     * @param array $timeouts
     * @param array $headers
* //     * @param int $retry
     * @param array $proxies
     * @param array $curl_opts
     * @return array
     */
    static function MULTI_CURL($urls = [], $methods = [], $posts = [], $timeouts = [], $headers = [], $proxies = [], $curl_opts = [])
    {
        //step one 初始化一个批处理handle
        $mh = curl_multi_init();

        //step two 往批处理handle 添加curl_init来的子handle
        foreach ($urls as $i => $url)
        {
            $ch = self::CH($url, $timeouts[$i], $headers[$i], $posts[$i], $proxies[$i], $methods[$i], $curl_opts);
            $conn[$i] = $ch;

            curl_multi_add_handle($mh, $conn[$i]);
        }

        //step three 并发执行，直到请求全部结束
        do {
            curl_multi_exec($mh, $active);
        } while ($active);

        $response = [];
        //step four 获取结果，移除并close handle
        foreach ($urls as $i => $url)
        {
            $response_content = curl_multi_getcontent($conn[$i]);
            $response[$i] =  self::createResponse($response_content, $conn[$i]);

            // 监控
            self::addMon(
                $url,
                curl_errno($conn[$i]) === 0 ? curl_getinfo($conn[$i], CURLINFO_HTTP_CODE) : -999,
                round(curl_getinfo($conn[$i], CURLINFO_TOTAL_TIME),4) * 1000
            );

            // 日志
            self::addLog($methods[$i], $url, $posts[$i], $response_content);

            curl_multi_remove_handle($mh,$conn[$i]);
            curl_close($conn[$i]);
        }

        curl_multi_close($mh);

        //step five 结果返回
        return $response;
    }

    /*
     * 监控打点，模块名默认为第三方服务url，指标为请求返回的HTTP_code与总响应时间
     */
    private static function addMon($url, $http_code, $request_time) {
        MonUtil::proxyMon($url, $http_code, '', $request_time);
    }

    private static function addLog($method, $urls, $params, $response)
    {
        //过滤无意义的日志
        $log = true;
        $excludes = config('http_util', 'log.exclude');
        if (!empty($excludes)) {
            foreach ($excludes as $exclude) {
                if (is_array($urls)) {
                    foreach ($urls as $url) {
                        if (strpos($url, $exclude) !== false) {
                            $log = false;
                            break;
                        }
                    }
                } else {
                    if (strpos($urls, $exclude) !== false) {
                        $log = false;
                        break;
                    }
                }
            }
        }
        if (!empty($log)) {
            //
            $params = self::change($params);
            FileLog::info($method, json_encode([
                'url' => $urls,
                'params' => $params,
                'response' => $response
                ]));
        }
    }

    /**
     * 过滤resource类型字段
     * @param array $arr
     * @return array
     */
    private static function change($arr)
    {
        foreach ($arr as $k => $v) {
            if (is_array($v)) {
                $arr[$k] = self::change($v);
            } else {
                if (strpos(gettype($v), 'resource') !== false) {
                    $arr[$k] = gettype($v);
                }
            }
        }
        return $arr;
    }

    /**
     * @param string $response_content 请求返回结果
     * @param resource $ch_handle cURL resource
     * @return array
     */
    private static function createResponse(string $response_content,  $ch_handle): array
    {
        return [
            'code'      => curl_errno($ch_handle) === 0 ? self::CODE_SUCCESS : self::CODE_FAIL,
            'response'  => empty(json_decode($response_content, true)) ? $response_content : json_decode($response_content, true),
            'http_code' => curl_getinfo($ch_handle, CURLINFO_HTTP_CODE),
            'msg'       => curl_error($ch_handle),
        ];
    }

    /**
     * 默认返回
     * @return array
     */
    private static function defaultResponse(): array
    {
        return [
            'code'      => self::CODE_FAIL,
            'response'  => '',
            'http_code' => 0,
            'msg'       => '',
        ];
    }
}
